local require = require

local system = require("common.system")
local pystring = require("pystring")
local dirent = require("posix.dirent")
local workVar = require("module.workVar")

local ipairs = ipairs
local pairs = pairs
local pcall = pcall
local format = string.format
local create = coroutine.create
local resume = coroutine.resume
local coReport = system.coReport
local msleep = workVar.msleep
local direntFiles = dirent.files
local pywith = pystring.with

local mt = {}

local function monProc(cb, proc)
    proc = proc or "/proc"
    local ok = true
    local tblRun  = {}  -- pid comm
    repeat
        local now = {}
        for pid in direntFiles(proc) do
            if pid:match("^[0-9]+$") then
                if not tblRun[pid] then
                    local path = format("%s/%s/comm", proc, pid)
                    local success, comm
                    success, comm = pcall(pywith, path)
                    if success and comm then
                        comm = comm:sub(1, -2)  -- strip last '\n'
                        tblRun[pid] = comm
                        cb(pid, comm)
                    end
                end
                now[pid] = ""
            end
        end

        for pid, _ in pairs(tblRun) do
            if not now[pid] then
                tblRun[pid] = nil
            end
        end
        msleep(3.3 * 1000)
    until not ok
end

local function cbNewproc(pid, comm)
    if comm:match("^ilogtail") then
        print("add mon: ", pid, comm)
    end
end

function mt.monProc(cb, path)
    cb = cb or cbNewproc
    local co = create(monProc)
    local ok, msg = resume(co, cb, path)
    coReport(co, ok, msg)
    return co
end


return mt