local require = require
require("eclass")
local flock = require("liveTrace.lib.flock")

local onperfFfi = require("liveTrace.modffi.nofpFfi")
local ffi, cffi = onperfFfi.ffi, onperfFfi.cffi
local decode_cffi = require("liveTrace.modffi.nfpdecodeFfi")
local dirent = require("posix.dirent")
local pystring = require("pystring")
local monProc = require("liveTrace.lib.monProc")

local livetrace_profiler_create = cffi.livetrace_profiler_create
local livetrace_profiler_destroy = cffi.livetrace_profiler_destroy
local livetrace_profiler_ctrl = cffi.livetrace_profiler_ctrl
local livetrace_enable_system_profiling = cffi.livetrace_enable_system_profiling
local livetrace_profiler_read_bytes = cffi.livetrace_profiler_read_bytes
local livetrace_disable_symbolizer = cffi.livetrace_disable_symbolizer

local pcall = pcall
local pairs = pairs
local error = error
local floor = math.floor
local time = os.time
local concat = table.concat
local io_open = io.open
local format = string.format
local match = string.match
local find = string.find
local c_cast = ffi.cast
local tostring = tostring
local split = pystring.split
local pywith = pystring.with
local rstrip = pystring.rstrip
local direntFiles = dirent.files
local c_string = ffi.string

local handle

local Cotel = class("otel")

local function ctrl_add()
    return function(pid, comm)
        if comm:match("^postgre") then
            -- print("add mon: ", pid, comm)
            livetrace_profiler_ctrl(handle, 1, tostring(pid))
        end
    end
end

local storeCells, prefix

function Cotel:_init_(args, cells)
    storeCells = cells
    prefix = "on"
    self._fdLock = flock.lock(args.lock)
    livetrace_disable_symbolizer()
    livetrace_enable_system_profiling()
    handle = livetrace_profiler_create()
    if not handle then
        error("create mullprof failed.")
    end
    monProc.monProc(ctrl_add())
end

local function nofp_live_cb(stream, size)
    local s = c_string(stream, size)
    storeCells[prefix] = s
end

local function read()
    local _cb = c_cast("nofp_live_cb", nofp_live_cb)
    livetrace_profiler_read_bytes(handle, _cb)
    _cb:free()
    return 0
end

function Cotel:pre()
end

function Cotel:calls()
    return read()
end

function Cotel:call()
    return read()
end

function Cotel:stop()
    livetrace_profiler_destroy(handle)
    handle = nil
    flock.unlock(self._fdLock)
end

return Cotel